import sys

import click
import questionary

import agent
from agent.server import start_server
from agent.function import Chat, LeaderAgent
from agent.utils import *
from agent.types import Project

console = Console()
configuration = Config()


def build_config(general: bool = False):
    """
    build_config: build the configuration for MLE-agent.
    Args:
        general: a boolean indicating whether to build the general configuration only.
    :return:
    """
    platform = questionary.select(
        "Which language model platform do you want to use?",
        choices=[LLM_TYPE_OPENAI, LLM_TYPE_OLLAMA]
    ).ask()

    # write the Ollama configuration
    if platform == LLM_TYPE_OLLAMA:
        model = questionary.text("What is the Ollama model version?").ask()

        if not model:
            sys.exit(0)

        platform_config = {
            "model": 'llama3'
        }

    # write the OpenAI configuration
    elif platform == LLM_TYPE_OPENAI:
        api_key = questionary.text("What is your OpenAI API key?").ask()

        if not api_key:
            sys.exit(0)

        platform_config = {
            "api_key": api_key,
            "model": 'gpt-3.5-turbo',
            'temperature': 0.7,
            'max_tokens': 16_385,
            'top_p': 1.0,
            'top_k': 32,
            'stop_sequences': 'None',
            'candidate_count': 1
        }
    else:
        console.log("The platform is not supported. Aborted.")
        sys.exit(0)

    search_engine = questionary.select(
        "Which search engine do you want to use?",
        choices=[SEARCH_ENGINE_GOOGLE, SEARCH_ENGINE_SEARCHAPI, SEARCH_ENGINE_BING, "no_web_search"]
    ).ask()

    if search_engine == "no_web_search":
        search_engine_config = {}
        console.log("No web search engine is set.")
    elif search_engine == SEARCH_ENGINE_GOOGLE:
        search_key = questionary.text("What is your Google search engine key?").ask()
        search_engine_id = questionary.text("What is your Google search engine ID?").ask()
        if not search_key or not search_engine_id:
            sys.exit(0)

        search_engine_config = {
            "key": search_key,
            "endpoint": "https://customsearch.googleapis.com/customsearch/v1",
            "cx": search_engine_id,
            "refer_count": 8,
            "timeout": 5
        }
    else:
        console.log(f"The {search_engine} search is not supported. Aborted.")
        sys.exit(0)

    code_language = CODE_LANGUAGE

    general_config = {
        'platform': platform,
        'code_language': code_language,
        'search_engine': search_engine,
    }

    configuration.write_section(CONFIG_SEC_GENERAL, general_config)
    if not general:
        configuration.write_section(platform, platform_config)
        configuration.write_section(search_engine, search_engine_config)


@click.group()
@click.version_option(version=agent.__version__)
def cli():
    """
    MLE-Agent: The CLI tool to build machine learning projects.
    """
    pass


@cli.group()
def project():
    """
    project: group of commands related to project management.
    """
    pass


@project.command()
@click.argument('name')
def delete(name):
    """
    delete: delete a machine learning project with the given NAME.
    """
    delete_project(name)


@project.command()
def switch():
    """
    switch: set the current project to work on.
    """
    projects = list_projects()
    project_names = [p.name for p in projects]
    target_name = questionary.select(
        "Select the project to work on:",
        choices=project_names
    ).ask()

    if target_name:
        target_project = read_project_state(target_name)
        configuration.write_section(
            CONFIG_SEC_PROJECT, {
                'path': target_project.path,
                'name': target_project.name
            }
        )
        console.log(f"Project set to: {target_name}")


@project.command()
def ls():
    """
    ls: list all existing machine learning projects.
    """
    projects = list_projects()
    for p in projects:
        if p.description:
            print(f"- {p.name}: {p.description}")
        else:
            print(f"- {p.name}")


@project.command()
def show():
    """
    show: show the current project.
    """
    if configuration.read().get('project') is None:
        console.log("You have not set up a project yet.")
        console.log("Please create a new project first using 'mle new' command.")
        return

    project_name = configuration.read()['project']['name']
    target_project = read_project_state(project_name)
    print(target_project.dict())


@cli.command()
@click.option('--general', '-g', is_flag=True, help="Set up the general configuration for MLE Agent.")
def config(general):
    """
    Set up the global configuration for MLE-agent.
    """
    build_config(general)


@cli.command()
def start():
    """
    start: start the chat with LLM.
    """
    if configuration.read() is None:
        build_config()

    if configuration.read().get('project') is None:
        console.log("You have not set up a project yet.")
        console.log("Please create a new project first using 'mle new your_project_name' command,"
                    " or set the project using 'mle set_project <project path>'.")
        return

    p = read_project_state(configuration.read()['project']['name'])
    if p is None:
        console.log("Could not find the project in the database. Aborted.")
        return

    if os.path.exists(p.path):
        chain = LeaderAgent(p, load_model())
        chain.start()
    else:
        console.log(f"Project path'{p.path}' does not exist. Aborted.")


@cli.command()
def chat():
    """
    chat: start an interactive chat with LLM to work on your ML project.
    """
    # check the system configuration
    if configuration.read() is None:
        build_config()

    model = load_model()
    console.log("Welcome to MLE-Agent! :sunglasses:")
    if model:
        console.log(f"Model loaded: {model.model_type}, {model.model}")
    console.line()

    if configuration.read().get('project') is None:
        console.log("You have not set up a project yet.")
        console.log("Please create a new project first using 'mle new' command.")
        return

    project_name = configuration.read()['project']['name']
    target_project = read_project_state(project_name)

    console.log("> [green]Current project:[/green]", target_project.path)
    console.log("> [green]Project language:[/green]", target_project.lang)

    current_task = target_project.plan.tasks[target_project.plan.current_task - 1]
    console.log("> [green]Current task:[/green]", current_task.name)
    console.log("> [green]Task progress:[/green]",
                f"{target_project.plan.current_task}/{len(target_project.plan.tasks)}")
    console.log("> [green]Task description:[/green]", current_task.description)

    # start the interactive chat
    console.line()
    chat_app = Chat(model)
    # set the initial system prompt
    chat_app.add(role='system', content=pmpt_chat_init(target_project))
    chat_app.start()


@cli.command()
@click.argument('name')
def new(name):
    """
    new: create a new machine learning project with the given NAME.
    """
    if not name:
        console.log("Please provide a valid project name. Aborted.")
        return

    description = questionary.text("What is the description of this project? (Optional)").ask()
    create_project(
        Project(
            name=name,
            description=description,
            lang=configuration.read()['general']['code_language'],
            llm=configuration.read()['general']['platform']
        ),
        set_current=True
    )


@cli.command()
@click.option('--port', '-p', default=8000, help="The port to run the server.")
@click.option('--address', '-a', default='0.0.0.0', help="The address to run the server.")
def server(port, address):
    """
    server: launch the MLE-Agent RESTFul API server.
    :param port: the target port.
    :param address: the target address.
    :return: None
    """
    start_server(address, port)


@cli.command()
def status():
    """
    status: display the current status of the project.
    """
    if configuration.read().get('project') is None:
        console.log("You have not set up a project yet.")
        console.log("Please create a new project first using 'mle new' command.")
        return

    project_name = configuration.read()['project'].get('name')
    if not project_name:
        console.log(
            "Currently no working project. "
            "Please use `mle project switch` to set a project or use `mle new` to create a new project."
        )
        return

    target_project = read_project_state(project_name)

    console.log("> [green]Current project:[/green]", target_project.name)
    console.log("> [green]Project path:[/green]", target_project.path)
    console.log("> [green]Project entry file:[/green]", target_project.entry_file)
    console.log("> [green]Project language:[/green]", target_project.lang)
    console.line()

    if target_project.plan:
        # display the current task name
        current_task = target_project.plan.tasks[target_project.plan.current_task - 1]
        console.log("> [green]Current task:[/green]", current_task.name)
        console.log("> [green]Task progress:[/green]",
                    f"{target_project.plan.current_task}/{len(target_project.plan.tasks)}")
        console.log("> [green]Task description:[/green]", current_task.description)

        if current_task.resources:
            console.line()
            console.log("> [green]Resources:[/green]")
            for resource in current_task.resources:
                console.log(f"- {resource.name}: {resource.uri}")

        if current_task.functions:
            console.line()
            console.log("> [green]Functions:[/green]")
            for function in current_task.functions:
                console.log(f"- {function.name}: {function.description}")

        if current_task.debug:
            console.line()
            console.log("> [green]Debugging:[/green]")
            console.log(f"- Maximum debug attempts: {current_task.debug}")
            console.log(f"- Debugging environment: {target_project.debug_env}")



import os

import pandas as pd


def read_csv_file(file_path, limit=3, column_only=False):
    """
    Reads a CSV file and returns a DataFrame with a limited number of rows or only column names.

    Parameters:
    - file_path (str): The path to the CSV file.
    - limit (int, optional): The number of rows to read from the CSV file. Defaults to None, which reads all rows.
    - column_only (bool, optional): If True, returns only the column names. Defaults to False.

    Returns:
    - DataFrame if the file exists and is read successfully and column_only is False.
    - List of column names if column_only is True.
    - None if the file does not exist.
    """
    if file_path is None:
        return None

    if os.path.exists(file_path):
        try:
            if column_only:
                df = pd.read_csv(file_path, nrows=1)
                return df.columns.tolist()
            elif limit is not None:
                return pd.read_csv(file_path, nrows=limit)
            else:
                return pd.read_csv(file_path)
        except Exception as e:
            print(f"Error reading the CSV file: {e}")
            return None
    else:
        return None



from agent.types import Project
from .system import read_file_to_string


def pmpt_chat_init(project: Project) -> str:
    current_task = project.plan.tasks[project.plan.current_task - 1]
    target_source_code = read_file_to_string(project.entry_file)
    if target_source_code is None:
        return f"""
        You are an Machine learning engineer, and you are currently working on an ML project using {project.lang}
         as the primary language. And now you should answer the user's questions based on the following information.
        
        USEFUL INFORMATION:
        
        - Project Language: {project.lang}
        - Your project plan: {project.plan.dict()}
        - The current task you are working on: {current_task.dict()}
        
        """
    else:
        return f"""
        You are an Machine learning engineer, and you are currently working on an ML project using {project.lang}
         as the primary language. And now you should answer the user's questions based on the following information.
        
        USEFUL INFORMATION:
        
        - Project Language: {project.lang}
        - Your project plan: {project.plan.dict()}
        - The current task you are working on: {current_task.dict()}
        - The source code you have written for the whole project: {read_file_to_string(project.entry_file)}
        
        """


def pmpt_public_dataset_guess():
    return f"""
    As an ML project expert, you provide consultation to users based on their specific requirements for datasets.

    Task:
    - Select three datasets that best meet the user's project requirements.
    - For each dataset, provide a one-sentence summary that explains why the dataset is appropriate for the project.

    Instructions:
    - Return the names of the datasets, each followed by a brief summary, formatted as entries in a list.
    - Each entry in the list should be a string containing the dataset name followed directly by a comma and the summary.

    Output Format:
    Your response should be formatted as a list of strings. Each string should follow this format:
    "Dataset_Name, Summary of why the dataset is suitable.",

    Example Output:
    [
        "TrafficVolumeDataset, Contains hourly traffic volume for various highways, ideal for predictive traffic
        modeling.",
        "RetailSalesForecasting, Provides daily sales data across several retail chains, suitable for demand
        forecasting.",
        "SentimentAnalysisTweets, Comprised of labeled tweets for sentiment analysis, perfect for training NLP models.",
    ]

    Ensure your response strictly adheres to this format for clarity and consistency.
    """


def pmpt_task_desc():
    return f"""
    You are an ML engineer that generates a task guide based on the task name,
     resources, and the user's requirements.

    - You should generate the task guide description based on the user's requirements.
    - You are welcome to add as more details as possible, but don't add irrelevant information under the task.
    - You should add an attribute 'description' to the task.

    EXAMPLE OUTPUT in JSON FORMAT:

    "tasks": [
    {{'name': 'Data Pre-processing', 'resources': [],
     'description': 'This task is first to preprocess the data. Second to ...'}},
    {{'name': 'Model Training', 'resources': 'PyTorch',
     'description': 'This task is to train the model using PyTorch, ...'}},
    {{'name': 'Model Deployment', 'resources': ['Flask'],
     'description': 'This task is to deploy the model using Flask, ...'}}
    ]

    """


def pmpt_plan(task_list):
    return f"""
    You are an machine learning engineer with strong expertise in MLOps and project planning.
    You will help me generate a project plan based on the user's requirements and provided information.
    
    - Understand the user's requirements and select the appropriate tasks.
    - Use the tasks only from the following task list.
    - Understand the task requirements and select the necessary resources.
    - Use appropriate resources for each task to track the experiment and make the project reproducible.
    - Select the minimum resources necessary to achieve the requirements.
    - Return the plan in JSON format, without any other information.

    AVAILABLE TASKS AND RESOURCES:
    {task_list}

    EXAMPLE OUTPUT in JSON FORMAT:

    {{
        "tasks": [
            {{"name": "Public Datasets Collection", "resources": ["HuggingFace Datasets"]}},
            {{"name": "Data Loading", "resources": ["pandas"]}},
            {{"name": "Data Pre-processing", "resources": ["scikit-learn preprocessing"]}},
            {{"name": "Model Training", "resources": ["huggingface transformers", "wandb"]}},
            {{"name": "Model Evaluation", "resources": ["torchmetrics"]}},
        ]
    }}
    """



import os
import sys
import subprocess
import importlib.util

from .base import BaseAgent
from .search_agent import SearchAgent
from agent.utils import Config, read_file_to_string

config = Config()


class ReflectAgent(BaseAgent):

    def gen_running_cmd(self):
        """
        Generate the running command for the current project.
        :return: the running command.
        """
        return f"python {os.path.basename(self.project.entry_file)}"

    def run_local(self):
        """
        Run a command in the shell, print the output and error logs in real time, and return the results.
        :return: A tuple containing the output, error, and exit status.
        """
        command = self.gen_running_cmd()
        os.chdir(self.project.path)
        output_log = ""
        error_log = ""
        try:
            process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            while True:
                output_line = process.stdout.readline()
                if output_line:
                    print(output_line, end='')
                    output_log += output_line

                error_line = process.stderr.readline()
                if error_line:
                    print(f"{error_line}", end='', file=sys.stderr)
                    error_log += error_line
                if output_line == '' and error_line == '' and process.poll() is not None:
                    break

            exit_code = process.wait()
            return output_log, error_log, exit_code
        except Exception as e:
            print(f"Exception occurred while running command: {command}\n Error: {str(e)}")
            return "", str(e), -1

    def run_cloud(self, cloud_type: str, dependency_list: list = None):
        """
        Debug the code script on the cloud service.
        :param cloud_type: the type of cloud service to use.
        :param dependency_list: the list of dependencies to install.
        :return:
        """

        dependency = "sky"
        spec = importlib.util.find_spec(dependency)
        if spec is not None:
            sky = importlib.import_module(dependency)
        else:
            raise ImportError(
                "It seems you didn't install the 'skypilot' package."
                f" Please install it using 'pip install skypilot-nightly[{cloud_type}]' command."
            )

        # check the workspace
        if not os.path.exists(self.project.path) or not os.path.isdir(self.project.path):
            self.console.log(f"The directory '{self.project.path}' does not exist or is not a directory.")
            return

        # generate the setup according to the dependency list.
        setup = ''
        if dependency_list:
            setup = f"pip install {' '.join(dependency_list)}"

        task_name = "task_" + os.path.basename(self.project.entry_file)
        task = sky.Task(
            name=task_name,
            setup=setup,
            workdir=self.project.path,
            run=self.gen_running_cmd()
        )

        def fetch_logs(job, status_only=False):
            """
            Check the status/logs of the job.
            :param job: the ID of a job.
            :param status_only: if True, return the status only.
            :return: the status of the job.
            """
            status = subprocess.run(
                f"sky logs --status {self.project.name} {job}",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True
            )

            code = -1
            status = status.stdout.split()[-1]
            if status == 'SUCCEEDED':
                code = 0
                if status_only:
                    return None, code
            else:
                if status_only:
                    return None, code

            logs = subprocess.run(
                f"sky logs {self.project.name} {job}",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True
            )
            return logs.stdout, code

        # set the cloud resource.
        if cloud_type == 'aws':
            task.set_resources(sky.Resources(cloud=sky.AWS()))
        elif cloud_type == 'gcp':
            task.set_resources(sky.Resources(cloud=sky.GCP()))
        elif cloud_type == 'azure':
            task.set_resources(sky.Resources(cloud=sky.Azure()))
        else:
            self.console.log(f"Cloud type '{cloud_type}' is not supported.")
            return

        job_id, _ = sky.launch(task, cluster_name=self.project.name)
        return fetch_logs(job_id)

    def run(self, cloud_type: str = None, dependency_list: list = None):
        """
        Run the code script.

        :param cloud_type: the type of cloud service to use, if None, run locally.
        :param dependency_list: the list of dependencies to install.
        """
        if cloud_type:
            run_log, exit_code = self.run_cloud(cloud_type, dependency_list)
        else:
            with self.console.status(f"Running the code script..."):
                run_log, error_log, exit_code = self.run_local()
        return run_log, exit_code

    def invoke(self, dependency_list=None, max_attempts: int = 3, cloud_type: str = None):
        """
        Run and debug the code script.

        :param cloud_type: the type of cloud service to use, if None, run locally.
        :param dependency_list: the list of dependencies to install.
        :param max_attempts: the max number of attempts to debug the code.
        :return:
        """
        pmpt_code_debug = f"""
        You are a Machine Learning engineer tasked with debugging a script.
        Your goal is to modify the code so that it meets the task requirements and runs successfully.
        You should modify the existing code based on the user's requirements, logs, and web search results.

        The output format should only with the updated code:

        Code: {{code}}
        """
        debug_success = False
        entry_file = self.project.entry_file
        run_log, exit_code = self.run(cloud_type, dependency_list)

        if exit_code != 0:
            enable_web_search = False if config.read().get('general').get('search_engine') == "no_web_search" else True
            search_agent = SearchAgent(enable_web_search)

            for attempt in range(max_attempts):
                self.console.log("Debugging the code script...")
                existing_code = read_file_to_string(entry_file)
                # TODO: summary the search keywords from the logs.
                search_results = search_agent.invoke(run_log)

                user_prompt = f"""
                Task Requirements: {self.project.requirement}\n
                Source Code: {existing_code}\n
                Log: {run_log}\n
                Web Search: {search_results}
                """

                self.chat_history.extend(
                    [
                        {"role": 'system', "content": pmpt_code_debug},
                        {"role": 'user', "content": user_prompt}
                    ]
                )

                self.handle_streaming()
                run_log, exit_code = self.run(cloud_type, dependency_list)

                if exit_code == 0:
                    debug_success = True
                    self.console.log("Debugging successful, the code script has been saved.")
                    break

            if not debug_success:
                if cloud_type:
                    subprocess.run(f"sky down {self.project.name}", shell=True)
                self.console.log(f"Debugging failed after {max_attempts} attempts.")
                return
        else:
            if cloud_type:
                subprocess.run(f"sky down {self.project.name}", shell=True)
            self.console.log("The code script has been run successfully.")
            return



import json
import questionary

from .base import BaseAgent
from agent.hub import load_yml
from agent.types import DebugEnv
from agent.function import ReflectAgent
from agent.utils import run_commands, preprocess_json_string, update_project_state, read_file_to_string


class SetupAgent(BaseAgent):
    """
    The LaunchAgent class.
    """

    def dependency_generator(self, code):
        """
        Generate the dependencies for the project plan.
        """
        pmpt_code_dependency = """
        As an ML project expert, your task is to analyze the user-provided code and detect the necessary dependencies
        required for the project. Based on this analysis, you will generate a list of shell commands that the user can
        execute to install these dependencies.

        Your output should include:
        - A list of detected dependencies.
        - Corresponding shell commands for installing these dependencies.
        - Ensure that these commands are suitable for execution in the user's environment.

        The output must be structured in JSON format to facilitate easy parsing and integration in different environments or tools.

        INPUT:
        Code provided by the user.

        EXAMPLE INPUT:
        import torch
        from transformers import BertModel

        EXPECTED OUTPUT in JSON FORMAT:
        {
            "dependencies": ["torch", "transformers", "build-essential"],
            "commands": ["python -m pip install torch", "pip install transformers", "apt-get install build-essential"]
        }
        """

        chat_history = [
            {"role": 'system', "content": pmpt_code_dependency},
            {"role": 'user', "content": code}
        ]
        return json.loads(preprocess_json_string(self.model.query(chat_history)))

    def invoke(self):
        """
        Invoke the agent.
        """
        # configurate the debug environment.
        self.project.debug_env = questionary.select(
            "Select the debug environment:",
            choices=[DebugEnv.not_running.value, DebugEnv.local.value, DebugEnv.cloud.value]
        ).ask()
        update_project_state(self.project)

        if self.project.debug_env == DebugEnv.not_running.value:
            self.console.log("The code execution and reflection are skipped.")
            return

        # read the source code and prepare the dependencies.
        source_code = read_file_to_string(self.project.entry_file)
        with self.console.status("Guessing and preparing the dependencies for the project plan..."):
            install_commands = self.dependency_generator(source_code).get('commands')
            dependencies = self.dependency_generator(source_code).get('dependencies')
            self.console.log(f"[cyan]Will install the following dependencies:[/cyan] {dependencies}")

        # confirm the installation for local debug environment.
        confirm_install = questionary.confirm("Are you sure to install the dependencies?").ask()
        if self.project.debug_env == DebugEnv.local.value:
            if confirm_install:
                run_commands(install_commands)
            else:
                self.console.log("Skipped the dependencies installation. You may need to install them manually.")
        elif self.project.debug_env == DebugEnv.cloud.value:
            if not confirm_install:
                dependencies = []
                self.console.log("Skipped the dependencies installation on the cloud.")

        cloud_type = None
        reflect_agent = ReflectAgent(self.model, self.project)
        if self.project.debug_env == DebugEnv.cloud.value:
            cloud_type = questionary.select("Select the cloud service:", choices=load_yml('cloud.yml')).ask()

        reflect_agent.invoke(cloud_type=cloud_type, dependency_list=dependencies)



from abc import ABC

from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel

from agent.types import Project
from agent.utils import extract_code, Config

config = Config()


class BaseAgent(ABC):
    def __init__(self, model, project: Project):
        self.model = model
        self.project = project
        self.requirement = self.project.requirement

        self.chat_history = []
        self.console = Console()
        self.project_home = config.read().get('project')['path']

    def handle_streaming(self):
        """
        Handle the streaming completion.
        :return: the result.
        """
        text = ""
        with Live(console=self.console) as live:
            for token in self.model.query(self.chat_history):
                if token:
                    text = text + token
                    live.update(
                        Panel(Markdown(text), title="[bold magenta]MLE-Agent[/]", border_style="magenta"),
                        refresh=True
                    )

            code = extract_code(text)
            if code:
                with open(self.project.entry_file, 'w') as file:
                    file.write(code)
                self.console.log(f"Code generated to: {self.project.entry_file}")
        return text



def download_readme(model_id):
    """
    Download the README of the specified model.

    Parameters:
    model_id (str): The model ID to download the README from.

    Returns:
    str: The README content of the specified model.
    """
    import huggingface_hub as hf
    try:
        file_path = hf.hf_hub_download(model_id, 'README.md')
        with open(file_path, 'r', encoding='utf-8') as file:
            content = file.read()
        return content
    except Exception as e:
        return f"An error occurred: {e}"


def get_top_downloaded_models(
        task_name: str,
        top_k=5,
        tags=None,
        language: str = None,
        library: str = None
):
    """
    Search for models based on the specified task and return the top-downloaded K model ids.

    Parameters:
    task_name (str): The task to search for (e.g., "text-classification", "translation").
    top_k (int): The number of top-downloaded model ids to return.
    tags (List, str): A string tag or a list of tags to filter models on the Hub by.
    language (str): A string or list of strings of languages, both by name and country code.
    library (str): A string or list of strings of foundational libraries models were originally trained from.

    Returns:
    list: A list of the top-downloaded model ids.
    """

    from huggingface_hub import HfApi, ModelFilter
    api = HfApi()

    try:
        models = list(
            api.list_models(
                filter=ModelFilter(
                    task=task_name,
                    tags=tags,
                    language=language,
                    library=library
                )
            )
        )

        if not models:
            return "No models found for the specified task."

        models_sorted = sorted(models, key=lambda x: x.downloads, reverse=True)
        return [model.id for model in models_sorted[:top_k]]
    except Exception as e:
        return f"An error occurred: {e}"



from .base import BaseAgent
from .chat import Chat
from .code_gen_agent import CodeAgent
from .reflect_agent import ReflectAgent
from .search_agent import SearchAgent
from .launch_agent import SetupAgent
from .tech_leader import LeaderAgent



from abc import ABC, abstractmethod


class Model(ABC):

    def __init__(self):
        self.model_type = None

    @abstractmethod
    def query(self, chat_history):
        pass

    @abstractmethod
    def stream(self, chat_history):
        pass



from .base import BaseAgent
from agent.types import Task
from agent.utils import read_file_to_string, update_project_state


class CodeAgent(BaseAgent):

    def system_pmpt_code_gen_new(self) -> str:
        return f"""
        You are a Machine Learning engineer working on an ML project using {self.project.lang} as the primary language.
        The user requires a new code script for the current task. Please read users requirement and task description
        very carefully and generate a new code script that meets them step by step.
        
        Please make sure the output only contains the source code.

        The output format should be:

        Code: {{code}}
        """

    def system_pmpt_code_gen_existing(self, code: str) -> str:
        return f"""
        You are a Machine Learning engineer working on an ML project using {self.project.lang} as the primary language.
        The user requires modifications to the existing code to meet the task requirements.
        
        Please use the following information to modify or add new changes to the existing code to finish the task.
        Please make sure the output only contains the source code.

        Existing Code: {code}

        The output format should be:

        Code: {{code}}
        """

    def task_prompt(self, task: Task, requirement) -> str:
        return f"""
        User Requirement: {requirement}
        Primary Language: {self.project.lang}
        Current Task: {task.name}
        Task Description: {task.description}
        """

    def gen_code(self, task: Task, requirement):
        """
        Generate the content of the current task.
        :param task: the task to work on
        :param requirement: the user's requirement

        :return: the content of the task.
        """
        entry_file = self.project.entry_file
        sys_prompt = self.system_pmpt_code_gen_new()
        if entry_file:
            existing_code = read_file_to_string(entry_file)
            if existing_code or self.project.plan.current_task <= 1:
                sys_prompt = self.system_pmpt_code_gen_existing(existing_code)
            else:
                self.console.log(
                    f"File {entry_file} not found. "
                    f"Please make sure the script exists or deleting the {entry_file} in the project.yml \n"
                )
                return None

        self.chat_history.extend(
            [
                {"role": 'system', "content": sys_prompt},
                {"role": 'user', "content": self.task_prompt(task, requirement)}
            ]
        )

        code = self.handle_streaming()
        return code

    def invoke(self, task_num, requirement):
        for task in self.project.plan.tasks:
            if self.project.plan.current_task < task_num:
                self.console.log(f"Working on task: {task.name} ({self.project.plan.current_task + 1}/{task_num})")
                # TODO: add supports for other kind of tasks.
                if task.kind == 'code_generation':
                    result = self.gen_code(task, requirement)
                    if result is None:
                        self.console.log("[red]Task failed. Aborting the chain.")
                        return

                self.project.plan.current_task += 1
            update_project_state(self.project)



import importlib.util

from agent.types.const import LLM_TYPE_OPENAI, LLM_TYPE_OLLAMA
from agent.model.base import Model


class OllamaModel(Model):
    def __init__(self, model: str, host_url=None):
        """
        Initialize the Ollama model.
        Args:
            host_url (str): The Ollama Host url.
            model (str): The model version.
        """
        super().__init__()

        dependency = "ollama"
        spec = importlib.util.find_spec(dependency)
        if spec is not None:
            self.model = model
            self.model_type = LLM_TYPE_OLLAMA
            self.ollama = importlib.import_module(dependency)
            self.client = self.ollama.Client(host=host_url)
        else:
            raise ImportError(
                "It seems you didn't install ollama. In order to enable the Ollama client related features, "
                "please make sure ollama Python package has been installed. "
                "More information, please refer to: https://github.com/ollama/ollama-python"
            )

    def query(self, chat_history):
        """
        Query the LLM model.
        Args:
            chat_history: The context (chat history).
        """
        return self.client.chat(model=self.model, messages=chat_history)['message']['content']

    def stream(self, chat_history):
        """
        Stream the output from the LLM model.
        Args:
            chat_history: The context (chat history).
        """
        for chunk in self.client.chat(
                model=self.model,
                messages=chat_history,
                stream=True
        ):
            yield chunk['message']['content']


class OpenAIModel(Model):
    def __init__(self, api_key, model, temperature):
        """
        Initialize the OpenAI model.
        Args:
            api_key (str): The OpenAI API key.
            model (str): The model with version.
            temperature (float): The temperature value.
        """
        super().__init__()

        dependency = "openai"
        spec = importlib.util.find_spec(dependency)
        if spec is not None:
            self.openai = importlib.import_module(dependency).OpenAI
        else:
            raise ImportError(
                "It seems you didn't install openai. In order to enable the OpenAI client related features, "
                "please make sure openai Python package has been installed. "
                "More information, please refer to: https://openai.com/product"
            )

        self.model = model
        self.model_type = LLM_TYPE_OPENAI
        self.temperature = temperature
        self.client = self.openai(api_key=api_key)

    def query(self, chat_history):
        """
        Query the LLM model.
        Args:
            chat_history: The context (chat history).
        """

        completion = self.client.chat.completions.create(
            model=self.model,
            messages=chat_history,
            temperature=self.temperature,
            stream=False
        )

        return completion.choices[0].message.content

    def stream(self, chat_history):
        """
        Stream the output from the LLM model.
        Args:
            chat_history: The context (chat history).
        """
        for chunk in self.client.chat.completions.create(
                model=self.model,
                messages=chat_history,
                temperature=self.temperature,
                stream=True
        ):
            yield chunk.choices[0].delta.content



import configparser
import os
from pathlib import Path

from agent.types.const import *

CONFIG_HOME = os.path.join(str(Path.home()), ".mle")
CONFIG_PATH = os.path.join(CONFIG_HOME, "config")


class Config:
    """
    Config: the overall system configuration for MLE-agent
    """

    def __init__(self):
        self.home = CONFIG_HOME
        Path(self.home).mkdir(parents=True, exist_ok=True)

        self.config_path = CONFIG_PATH
        self.config = configparser.ConfigParser()
        self.config.read(self.config_path)

    def read(self):
        """
        read: read the configuration file.

        Returns: a dictionary of the configuration.
        """
        if os.path.isfile(CONFIG_PATH):
            self.reload_config(CONFIG_PATH)
        else:
            return None

        config_dict = {}
        for section in self.config.sections():
            options_dict = {option: self.config.get(section, option) for option in self.config.options(section)}
            config_dict[section] = options_dict

        return config_dict

    def reload_config(self, config_path):
        """
        reload_config: The default configuration will load ~/.mle/config, if user want to specify
        customize, the method is required.

        @param config_path: the path of new configuration file.
        """
        self.config.read(config_path)

    def load_openai_config(self):
        """
        load_openai_config: load a OpenAI configuration when required.
        """
        if self.config.has_section(LLM_TYPE_OPENAI):
            return self.config[LLM_TYPE_OPENAI]
        else:
            raise ValueError("there is no '[openai]' section found in the configuration file.")

    def write_section(self, section_name: str, config_dict: dict, overwrite: bool = True):
        """
        write_section: write the project configuration.

        @param section_name: the section name.
        @param config_dict: the configuration dictionary.
        @param overwrite: a boolean indicating whether to overwrite the existing configuration.

        """
        if not self.config.has_section(section_name):
            self.config.add_section(section_name)

        # save the new configuration and reload.
        with open(self.config_path, 'w') as configfile:
            if overwrite:
                self.config[section_name] = config_dict
                self.config.write(configfile)
                self.reload_config(self.config_path)
            else:
                existing_config = dict(self.read().get(section_name))
                existing_config.update({key: value for key, value in config_dict.items() if key not in existing_config})
                self.config[section_name] = existing_config
                self.config.write(configfile)
                self.reload_config(self.config_path)



"""
The RESTFul server of MLE-Agent
"""
import uvicorn
from fastapi import FastAPI, HTTPException
from fastapi.responses import StreamingResponse
from fastapi.middleware.cors import CORSMiddleware

from agent.utils import *
from agent.function import Chat
from agent.types import Project, ConfigUpdateRequest

app = FastAPI()
origins = ["*"]
config = Config()

# handle the CROSS origin requests
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/")
def read_root():
    return {"message": "Welcome to MLE-Agent!"}


@app.put("/config/update")
def update_config(request: ConfigUpdateRequest):
    try:
        config.write_section(request.section_name, request.config_dict, request.overwrite)
        return {"message": "Configuration updated successfully"}
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@app.post("/projects/", response_model=Project)
def create_project(project: Project):
    create_project(project)
    return project


@app.get("/project/")
def list_project():
    """
    List all the projects.
    :return: the list of projects.
    """
    return {'projects': [p.dict() for p in list_projects()]}


@app.get("/project/{project_name}", response_model=Project)
def get_project(project_name: str):
    result = read_project_state(project_name)
    if result:
        return result
    else:
        raise HTTPException(status_code=404, detail="Project not found")


# run
@app.get("/run/")
def run_project():
    """
    run_project: Run the project.
    :return: the run status.
    """
    return {"status": "Running"}


@app.get("/chat/")
def chat(project: str, message: str):
    """
    Call the Chat's api function and return the streaming response. The input is an object with the project.
    """
    project_state = read_project_state(project)
    user_pmpt = message

    chat_app = Chat(load_model())
    chat_app.add(role='system', content=pmpt_chat_init(project_state))
    local_files_info = list_all_files(project_state.path)
    chat_app.add("user", f"""The files under the project directory is: {local_files_info}""")

    def generate_response(p):
        previous_text = ''
        for text in chat_app.handle_response(p):
            new_text = text[len(previous_text):]
            previous_text = text
            yield new_text

    return StreamingResponse(generate_response(user_pmpt), media_type="text/event-stream")


def start_server(address: str = "0.0.0.0", port: int = 8080):
    """
    Start the server.
    :param address: the server address.
    :param port: the server port.
    :return: None
    """
    uvicorn.run(app, host=address, port=port)



from enum import Enum
from pydantic import BaseModel
from typing import List, Optional


class DebugEnv(str, Enum):
    local = "local"
    cloud = "cloud"
    not_running = "not_running"


class SearchEngine(BaseModel):
    name: str
    endpoint: str
    key: str
    cx: Optional[str] = None
    refer_count: int = 8
    timeout: int = 5


class Resource(BaseModel):
    name: str
    uri: Optional[str] = None
    description: Optional[str] = None
    choices: Optional[List[str]] = None


class Function(BaseModel):
    name: str
    description: Optional[str] = None


class Task(BaseModel):
    name: str
    kind: str
    description: Optional[str] = None
    resources: Optional[List[Resource]] = None
    functions: Optional[List[str]] = None
    debug: Optional[int] = None


class Plan(BaseModel):
    current_task: int
    dataset: Optional[str] = None
    data_kind: Optional[str] = None
    ml_task_type: Optional[str] = None
    ml_model_arch: Optional[str] = None
    tasks: Optional[List[Task]] = None


class Project(BaseModel):
    name: str
    lang: str
    llm: str
    path: Optional[str] = None
    plan: Optional[Plan] = None
    entry_file: Optional[str] = None
    debug_env: Optional[str] = DebugEnv.local
    description: Optional[str] = None
    requirement: Optional[str] = None
    enhanced_requirement: Optional[str] = None



import requests
from fastapi import HTTPException
from rich.console import Console

from agent.types import SearchEngine
from agent.utils import Config

config = Config()


class SearchAgent:
    def __init__(self, enable_web_search: bool = False):
        self.enable_web_search = enable_web_search
        self.console = Console()

        if not self.enable_web_search:
            self.console.log("Web search is disabled.")
            return

        else:
            config_dict = config.read()
            self.engine_name = config_dict['general'].get('search_engine')
            if not self.engine_name:
                self.console.log("No search engine is set.")
                return
            search_engine = config.read().get(self.engine_name)
            search_engine['name'] = self.engine_name

            self.search_engine = SearchEngine.validate(search_engine)

    def search_with_google(self, query: str):
        """
        https://developers.google.com/custom-search/v1/using_rest
        """
        params = {
            "key": self.search_engine.key,
            "cx": self.search_engine.cx,
            "q": query,
            "num": self.search_engine.refer_count
        }
        self.console.log(f"Searching with Google")
        response = requests.get(
            self.search_engine.endpoint, params=params, timeout=self.search_engine.timeout
        )
        if not response.ok:
            self.console.log(f"{response.status_code} {response.text}")
            raise HTTPException(response.status_code, "Search engine error.")
        json_content = response.json()
        try:
            contexts = json_content["items"][:self.search_engine.refer_count]
        except KeyError:
            self.console.log(f"Error encountered: {json_content}")
            return []
        return contexts

    def invoke(self, query: str):
        """
        Invoke the search agent.
        """
        if not self.enable_web_search:
            return ["no web search results for this query."]

        if self.engine_name == "google":
            return self.search_with_google(query)
        else:
            self.console.log(f"Search engine {self.engine_name} is not supported.")
            raise HTTPException(400, "Search engine not supported.")



import ast

import questionary
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory

from agent.hub.utils import match_plan
from agent.integration import read_csv_file
from agent.types import Plan
from agent.utils import *
from .code_gen_agent import CodeAgent
from .plan_agent import (
    plan_generator,
    analyze_requirement,
    gen_file_name,
    pmpt_dataset_detect,
    pmpt_task_select,
    pmpt_model_select
)
from .launch_agent import SetupAgent

config = Config()


class LeaderAgent:
    def __init__(self, project: Project, model):
        """
        LeaderAgent: the interactive chain of the current ML task.
        :param project: the current working project.
        :param model: the language model.
        """
        self.project = project
        self.model = model
        self.chat_history = []
        self.console = Console()
        # if the project is not set up, then raise an error.
        if config.read().get('project') is None:
            self.console.log("You have not set up a project yet.")
            self.console.log("Please create a new project first using 'mle new project_name' command then try again.")
            raise SystemExit

        self.project_home = config.read().get('project')['path']
        self.session = PromptSession(
            history=FileHistory(str(os.path.join(self.project_home, CONFIG_TASK_HISTORY_FILE)))
        )

        self.entry_file = self.project.entry_file
        self.requirement = self.project.requirement
        # initialize the plan if it is not set up.
        if self.project.plan is None:
            self.project.plan = Plan(current_task=0)

    def user_requirement_understanding(self):
        """
        (STEP-0) User Requirement Understanding.
        :return:
        """
        if self.project.requirement:
            self.console.print(f"[cyan]User Requirement:[/cyan] {self.project.requirement}")
        else:
            show_panel("STEP 0: User Requirements Understanding")
            self.requirement = questionary.text("Hi, what are your requirements?").ask()
            self.project.requirement = self.requirement

        if not self.requirement:
            raise SystemExit("The user requirement is not provided.")

        # Generate entry file name based on requirement
        self.project.enhanced_requirement = self.requirement
        if self.entry_file is None:
            self.entry_file = gen_file_name(self.project, self.model)
        self.console.print(f"[cyan]Entry File:[/cyan] {self.entry_file}")
        update_project_state(self.project)

    def dataset_selection(self):
        """
        (STEP-1) Dataset Selection.
        :return:
        """

        if self.project.plan.data_kind is None and self.project.plan.dataset is None:
            show_panel("STEP 1: Dataset Selection")
            self.project.plan.data_kind = analyze_requirement(
                self.project.enhanced_requirement,
                pmpt_dataset_detect(),
                self.model
            )

            if self.project.plan.data_kind == 'no_data_information_provided':
                public_dataset_list = analyze_requirement(
                    self.project.enhanced_requirement,
                    pmpt_public_dataset_guess(),
                    self.model
                )
                public_dataset_list = ast.literal_eval(public_dataset_list)
                self.project.plan.dataset = questionary.select(
                    "Please select the dataset:",
                    choices=public_dataset_list
                ).ask()
            elif self.project.plan.data_kind == 'csv_data':
                self.project.plan.dataset = questionary.text("Please provide the CSV data path:").ask()
                if os.path.exists(self.project.plan.dataset) is False:
                    public_dataset_list = analyze_requirement(
                        self.project.enhanced_requirement,
                        pmpt_public_dataset_guess(),
                        self.model
                    )
                    public_dataset_list = ast.literal_eval(public_dataset_list)
                    self.project.plan.dataset = questionary.select(
                        "Please select the dataset:",
                        choices=public_dataset_list
                    ).ask()

        if self.project.plan.dataset is None:
            raise SystemExit("There is no dataset information. Aborted.")
        else:
            self.console.print(f"[cyan]Data Source:[/cyan] {self.project.plan.dataset}")
            if self.project.plan.data_kind == 'csv_data':
                csv_data_sample = read_csv_file(self.project.plan.dataset)
                self.console.print(f"[cyan]Dataset examples:[/cyan] {csv_data_sample}")
                self.project.enhanced_requirement += f"\nDataset Sample: {csv_data_sample}"

            self.project.enhanced_requirement += f"\nDataset: {self.project.plan.dataset}"
            update_project_state(self.project)

    def task_model_selection(self):
        """
        (STEP-2) Task & Model Selection.
        :return:
        """
        if self.project.plan.ml_task_type is None or self.project.plan.ml_model_arch is None:
            show_panel("STEP 2: Task & Model Selection")

        # select the ml task type
        if self.project.plan.ml_task_type is None:
            ml_task_list = analyze_requirement(self.project.enhanced_requirement, pmpt_task_select(), self.model)
            ml_task_list = ast.literal_eval(ml_task_list)
            ml_task_type = questionary.select(
                "Please select the ML task type:",
                choices=ml_task_list
            ).ask()

            self.console.log(f"[cyan]ML task type detected:[/cyan] {ml_task_type}")
            confirm_ml_task_type = questionary.confirm("Are you sure to use this ml task type?").ask()
            if confirm_ml_task_type:
                self.project.plan.ml_task_type = ml_task_type
            else:
                self.console.log("Seems you are not satisfied with the task type. Aborting the chain.")
                return

            self.project.enhanced_requirement += f"\n\nML task type: {self.project.plan.ml_task_type}"
        else:
            self.console.print(f"[cyan]ML Task:[/cyan] {self.project.plan.ml_task_type}")

        # select the mode architecture
        if self.project.plan.ml_model_arch is None:
            ml_model_list = analyze_requirement(self.project.enhanced_requirement, pmpt_model_select(), self.model)
            ml_model_list = ast.literal_eval(ml_model_list)
            ml_model_arch = questionary.select(
                "Please select the ML model architecture:",
                choices=ml_model_list
            ).ask()
            self.console.log(f"[cyan]Model architecture detected:[/cyan] {ml_model_arch}")
            confirm_ml_model_arch = questionary.confirm("Are you sure to use this ml arch?").ask()
            if confirm_ml_model_arch:
                self.project.plan.ml_model_arch = ml_model_arch
            else:
                self.console.log("Seems you are not satisfied with the model architecture. Aborting the chain.")
                return

            self.project.enhanced_requirement += f"\nModel architecture: {self.project.plan.ml_model_arch}"
        else:
            self.console.print(f"[cyan]Model Architecture:[/cyan] {self.project.plan.ml_model_arch}")

        update_project_state(self.project)

    def task_planning(self):
        """
        (STEP-3) Task Planning.
        :return:
        """
        if self.project.plan.tasks is None:
            show_panel("STEP 3: Task Planning")
            self.console.log(
                f"The project [cyan]{self.project.name}[/cyan] has no existing plans. Start planning..."
            )
            self.project.enhanced_requirement += f"\nDataset: {self.project.plan.dataset}"
            with self.console.status("Planning the tasks for you..."):
                task_dicts = plan_generator(self.project.enhanced_requirement, self.model)
                self.console.print(generate_plan_card_ascii(task_dicts), highlight=False)
                self.project.plan.tasks = []
                for task_dict in task_dicts.get('tasks'):
                    task = match_plan(task_dict)
                    if task:
                        self.project.plan.tasks.append(task)

            # Confirm the plan
            confirm_plan = questionary.confirm("Are you sure to use this plan?").ask()
            if confirm_plan:
                update_project_state(self.project)
            else:
                self.console.log("Seems you are not satisfied with the plan. Aborting the chain.")
                return
        else:
            tasks = []
            for t in self.project.plan.tasks:
                tasks.append({'name': t.name, 'resources': [r.name for r in t.resources], 'description': t.description})
            self.console.print(f"[cyan]Tasks:[/cyan]")
            self.console.print(generate_plan_card_ascii({'tasks': tasks}), highlight=False)

    def code_generation(self):
        """
        (STEP-4) Code Generation.
        :return:
        """
        task_num = len(self.project.plan.tasks)
        if self.project.plan.current_task < task_num:
            show_panel("STEP 4: Code Generation")
            code_generation_agent = CodeAgent(self.model, self.project)
            code_generation_agent.invoke(task_num, self.requirement)
            update_project_state(self.project)

    def execution_and_reflection(self):
        """
        (STEP-5) Execution and Reflection.
        :return:
        """
        show_panel("STEP 5: Execution and Reflection")
        launch_agent = SetupAgent(self.model, self.project)
        launch_agent.invoke()
        update_project_state(self.project)

    def start(self):
        """
        Execute the chain.
        :return: the result of the chain.
        """
        try:
            # STEP 0: User Requirement Understanding
            self.user_requirement_understanding()
            # STEP 1: Dataset Selection
            self.dataset_selection()
            # STEP 2: Task & Model Selection
            self.task_model_selection()
            # STEP 3: Task Planning
            self.task_planning()
            # STEP 4: Code Generation
            self.code_generation()
            # STEP 5: Execution and Reflection
            self.execution_and_reflection()
            self.console.log("The chain has been completed.")
        except KeyboardInterrupt:
            self.console.log("MLE Plan Agent has been interrupted.")
            return



import os
import re
import shutil
import subprocess
from pathlib import Path
from typing import TypeVar

from tinydb import TinyDB, Query
from pydantic import BaseModel
from rich.console import Console

from agent.types import Project
from agent.utils import Config, CONFIG_SEC_PROJECT
from agent.model import OpenAIModel, OllamaModel
from agent.types.const import TABLE_PROJECTS, LLM_TYPE_OPENAI, CONFIG_SEC_API_KEY, LLM_TYPE_OLLAMA

T = TypeVar('T', bound=BaseModel)
project_db = TinyDB(os.path.join(Config().home, TABLE_PROJECTS))


def load_model():
    """
    load_model: load the model based on the configuration.
    """
    config = Config()
    config_dict = config.read()
    plat = config_dict['general']['platform']

    model = None
    if plat == LLM_TYPE_OPENAI:
        model = OpenAIModel(
            api_key=config_dict[LLM_TYPE_OPENAI][CONFIG_SEC_API_KEY],
            model=config_dict[LLM_TYPE_OPENAI].get('model'),
            temperature=float(config_dict[LLM_TYPE_OPENAI]['temperature'])
        )
    if plat == LLM_TYPE_OLLAMA:
        model = OllamaModel(model=config_dict[LLM_TYPE_OLLAMA].get('model'))

    return model


def preprocess_json_string(json_string):
    """
    Preprocess a JSON string to handle single quotes and other special cases.
    :param json_string: the input JSON string.
    :return: the preprocessed JSON string.
    """
    # Replace single quotes with double quotes
    json_string = re.sub(r"'", r'"', json_string)

    # Handle cases where single quotes are used inside double quotes
    json_string = re.sub(r'"\s*:\s*"', r'": "', json_string)
    json_string = re.sub(r'"\s*,\s*"', r'", "', json_string)
    json_string = re.sub(r'\[\s*"', r'["', json_string)
    json_string = re.sub(r'"\s*\]', r'"]', json_string)

    return json_string


def list_all_files(path):
    """
    Lists all files and directories under the given path if it is a directory.
    If the path is a file, returns None.

    Args:
    path (str): The file system path to check and list contents from.

    Returns:
    list of str or None: A list of file and directory names under the given path, or None if the path is a file.
    """
    if os.path.isfile(path):
        return None  # Return None if the path is a file

    file_list = []
    for root, dirs, files in os.walk(path):
        for name in files:
            file_list.append(os.path.join(root, name))
        for name in dirs:
            file_list.append(os.path.join(root, name))

    return file_list


def create_directory(dir_name: str):
    """
    Create a directory if it does not exist.
    :param dir_name: the name of the directory to create.
    :return: the directory created.
    """
    console = Console()
    cwd = os.getcwd()
    path = os.path.join(cwd, dir_name)

    try:
        os.makedirs(path, exist_ok=True)
        console.log(f"[green]Directory '{path}' created successfully. Use 'mle start' to start the project.")
    except OSError as error:
        console.log(f"[red]Creation of the directory '{path}' failed due to: {error}")
    return path


def delete_directory(path):
    """
    delete_directory: delete a directory and all its contents.

    Args:
        path: The path to the directory to be deleted.
    """
    if os.path.exists(path):
        shutil.rmtree(path)
        return True
    else:
        return False


def get_directory_name(path):
    """
    Get the directory name if the path is a directory.
    :param path: the path to check.
    :return: the directory name if it's a directory, otherwise None.
    """
    if os.path.isdir(path):
        return os.path.basename(path)
    else:
        return None


def read_project_state(project_name: str):
    """
    Read the project state.
    :param project_name: the project name.
    :return: the project object.
    """
    query = Query()
    project = project_db.get(query.name == project_name)
    if project:
        return Project(**project)

    return None


def update_project_state(project: Project):
    """
    Update the project state.
    :param project: the project object.
    :return: the project object.
    """
    query = Query()
    if project_db.contains(query.name == project.name):
        project_db.update(project.dict(), query.name == project.name)
    else:
        project_db.insert(project.dict())


def delete_project(project_name: str):
    """
    Delete a project from the database.
    :param project_name: the name of the project to delete.
    :return: None
    """
    console = Console()
    p = read_project_state(project_name)
    if p is None:
        console.log(f"Project '{project_name}' not found.")
        return
    else:
        # delete the database record
        query = Query()
        project_db.remove(query.name == project_name)
        # delete the workspace
        delete_directory(p.path)
        # modify the configuration file
        config = Config()
        if config.read().get('project') and config.read()['project']['name'] == project_name:
            config.write_section(CONFIG_SEC_PROJECT, {})

        console.log(f"Project '{project_name}' deleted successfully.")


def create_project(project: Project, set_current=False):
    """
    Create a new project.
    :param project: the project object.
    :param set_current: the flag to set the current project.
    :return: the data of the new project.
    """
    cwd = os.getcwd()
    project_path = os.path.join(cwd, project.name)
    Path(project_path).mkdir(parents=True, exist_ok=True)

    project.path = project_path
    update_project_state(project)

    if set_current:
        Config().write_section(
            CONFIG_SEC_PROJECT, {
                'path': project_path,
                'name': project.name
            }
        )
    return project


def list_projects():
    """
    List all the projects.
    :return: the list of projects.
    """
    return [Project(**item) for item in project_db.all()]


def extract_and_save_file(input_text):
    """
    Extracts the file name and code block from a text formatted as specified,
    then creates a local file with the file name and writes the code into it.

    Args:
    input_text (str): The text containing the file name and code block.

    Returns:
    str: The name of the file created.
    """
    console = Console()
    file_name_match = re.search(r"File Name:\s*(.*?)\s*\n+\s*Code:", input_text, re.DOTALL)
    if not file_name_match:
        console.log("File name not found in the text.")
        return None, None

    file_name = file_name_match.group(1).strip()

    project_file_path = os.path.join(Config().read()['project']['path'], file_name)
    code_match = re.search(r"```(?:[a-zA-Z0-9]+)?\n(.*?)```", input_text, re.DOTALL)
    if not code_match:
        console.log("Code block not found in the text.")
        return None, None

    code = code_match.group(1).strip()
    with open(project_file_path, 'w') as file:
        file.write(code)

    return project_file_path, code


def extract_file_name(text: str):
    """
    Extracts the file name from a given text string.

    Args:
    text (str): The text containing the file name.

    Returns:
    str: The extracted file name, or an empty string if no file name is found.
    """
    match = re.search(r'File Name: (\S+)', text)
    if match:
        return match.group(1)
    else:
        return None


def extract_code(text: str):
    """
    Extracts the code block from a given text string.

    Args:
    text (str): The text containing the code block.

    Returns:
    str: The extracted code block, or an empty string if no code block is found.
    """
    match = re.search(r'```(?:[a-zA-Z0-9]+)?\n(.*?)```', text, re.DOTALL)
    if match:
        return match.group(1)
    else:
        return None


def read_file_to_string(file_path: str):
    """
    Reads the contents of a file and returns it as a string.

    Args:
    file_path (str): The path to the file that needs to be read.

    Returns:
    str: The contents of the file as a string.
    """
    try:
        with open(file_path, 'r', encoding='utf-8') as file:
            return file.read()
    except FileNotFoundError:
        return None


def run_commands(commands):
    """
    Run multiple commands in the shell and return the outputs, errors, and exit statuses.
    :param commands: the list of input commands to run.
    :return: a list of tuples containing the output, error (if any), and exit status for each command.
    """
    results = []
    for command in commands:
        try:
            process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

            output = ''
            while True:
                line = process.stdout.readline()
                if not line and process.poll() is not None:
                    break
                output += line
                print(line, end='')

            exit_code = process.wait()
            results.append((output, exit_code))
        except Exception as e:
            results.append((str(e), -1))


def list_dir_structure(start_path):
    """
    List all files and directories under the given path.
    :param start_path: the path to start listing from.
    :return:
    """
    return_str = ""
    for root, dirs, files in os.walk(start_path):
        level = root.replace(start_path, '').count(os.sep)
        indent = ' ' * 4 * level
        subindent = ' ' * 4 * (level + 1)
        return_str += f'{indent}{os.path.basename(root)}/\n'
        for f in files:
            return_str += f'{subindent}{f}\n'

    return return_str



import json
import os

import questionary
from rich.console import Console

from agent.hub import load_yml
from agent.utils import update_project_state
from agent.utils.prompt import pmpt_task_desc, pmpt_plan


def pmpt_plan_filename(lang: str) -> str:
    file_extension = {
        'Python': '.py',
        'JavaScript': '.js',
        'Java': '.java',
        # Add other languages and their extensions as needed
    }.get(lang, '.py')  # Default to .txt if the language is not listed

    return f"""
    You are a Machine Learning Engineer working on a project that primarily uses {lang}.
    Your task is to generate 5 file names based on the given user requirements.
    Ensure that the file suffix is correct for the specified language, which is '{file_extension}' for {lang}.

    Please generate 5 potential file names to allow the user to choose by following the example formats.

    Example output:
    ['train{file_extension}', 'data_loading{file_extension}', 'test_cases{file_extension}',
    'model_serve{file_extension}', 'utils{file_extension}']
    """


def gen_file_name(project, llm_agent):
    console = Console()
    prompt_message = pmpt_plan_filename(project.lang)

    with console.status("Preparing entry file name..."):
        chat_history = [
            {"role": 'system', "content": prompt_message},
            {"role": 'user', "content": project.requirement}
        ]
        file_name_candidates = llm_agent.query(chat_history)

        # If the output is a single string containing list representation, convert it properly
        if isinstance(file_name_candidates, str) and file_name_candidates.startswith(
                '[') and file_name_candidates.endswith(']'):
            import ast
            try:
                file_name_candidates = ast.literal_eval(file_name_candidates)
            except SyntaxError as e:
                console.log(f"Error parsing filenames: {e}")
                file_name_candidates = []  # Fallback to an empty list if parsing fails
        else:
            file_name_candidates = (file_name_candidates,)

        file_path_candidates = [os.path.join(project.path, filename.strip("'")) for filename in file_name_candidates]

    entry_file = questionary.select("Please select the file name:", choices=file_path_candidates).ask()
    # Update the project entry file and clear the chat history
    project.entry_file = entry_file
    update_project_state(project)

    return entry_file


def pmpt_dataset_detect():
    # Load and format the dataset names for display
    data_options = load_yml('data.yml')

    return f"""
    You are an machine learning engineer tasked with detecting the appropriate dataset based on user requirements.

    Instructions:
    - Detect the data source based on the user's specific requirements.
    - If the requirement does not specify a data source, use the default response: 'no_data_information_provided'.

    Available Data Sources:
    {data_options}

    Output:
    - Provide only the name of the dataset from the list, without any punctuation or additional formatting.
    """


def pmpt_task_select():
    task_list = load_yml('task.yml')

    return f"""
    As an ML project expert, you are tasked with selecting the most appropriate tasks based on the user's requirements
    and provided data samples.

    Instructions:
    - Analyze the provided user requirements and data samples carefully.
    - Select three tasks that are most relevant to the requirements.
    - For each selected task, provide a one-sentence description explaining its relevance.

    Available Tasks:
    {task_list}

    Output Format:
    Your response should include three entries formatted as a list of strings, where each string contains the task name followed by its description, e.g.:
    ["Task1: Description of Task1", "Task2: Description of Task2", "Task3: Description of Task3"]

    Note: Return only the task names followed by a brief description, without any additional information or punctuation.
    """


def pmpt_model_select():
    return f"""
    As an ML project expert, you are tasked with recommending three different AI model architectures based on the
    user's requirements, provided data samples, and selected ML tasks. Each recommended model should align with one
    of the specific criteria: best overall performance, balanced performance, and fastest computation.

    Instructions:
    - Analyze the provided user requirements, data samples, and ML tasks.
    - Select three models:
        1. The best performing model in terms of accuracy.
        2. A model with balanced performance considering both accuracy and computational efficiency.
        3. The fastest model in terms of computational speed.
    - For each model, provide a brief one-sentence summary that includes its suitability for the targeted task, its
    accuracy, and its speed.

    Output Format:
    Please return a list of three strings, where each string includes the model's name followed by its summary.
    Ensure the description highlights how the model meets one of the selection criteria (best, balanced, fastest).
    Example format:
    ["ModelName1: Best for task X with high accuracy of Y%, suitable for complex data analysis.",
     "ModelName2: Balanced model, offers moderate accuracy with better speed, good for real-time applications.",
     "ModelName3: Fastest model with lower accuracy, best for quick processing where speed is prioritized over precision."]

    Note: Ensure that the architecture names with summary are returned without any additional punctuation.
    """


def analyze_requirement(requirement: str, sys_prompt: str, llm_agent):
    """
    Generate the project plan based on the user's requirements.
    :param requirement: the user's requirements.
    :param sys_prompt: the system prompt.
    :param llm_agent: the language model agent.
    :return: the project plan.
    """
    chat_history = [
        {"role": 'system', "content": sys_prompt},
        {"role": 'user', "content": requirement}
    ]
    return llm_agent.query(chat_history)


def description_generator(requirement: str, task_list, llm_agent):
    """
    Generate the detailed description of the plan.
    :param requirement: the user's requirements.
    :param task_list: the selected task list.
    :param llm_agent: the language model agent.
    :return: the detailed description of the plan.
    """
    user_prompt = f"""
    Overall Requirements: {requirement}
    Tasks: {task_list}
    """

    chat_history = [
        {"role": 'system', "content": pmpt_task_desc()},
        {"role": 'user', "content": user_prompt}
    ]
    return json.loads(llm_agent.query(chat_history))


def plan_generator(
        requirement: str,
        llm_agent,
):
    """
    Generate the project plan based on the user's requirements.
    :param requirement: the user's requirements.
    :param llm_agent: the language model agent.
    :return: the project plan.
    """
    task_list = []
    for task in load_yml('plan.yml'):
        task_list.append(
            {
                'name': task['name'],
                'prompt': task['prompt'],
                'resources': [r['name'] for r in task.get('resources')]
            }
        )
    chat_history = [
        {"role": 'system', "content": pmpt_plan(json.dumps(task_list))},
        {"role": 'user', "content": requirement}
    ]
    target_tasks = llm_agent.query(chat_history)

    # generate the completed plan with detailed description
    return description_generator(requirement, target_tasks, llm_agent)



# Main
CODE_LANGUAGE = 'python'
# Configurations
CONFIG_SEC_GENERAL = 'general'
CONFIG_SEC_PROJECT = 'project'
CONFIG_SEC_API_KEY = 'api_key'
CONFIG_PROJECT_FILE = 'project.yml'
CONFIG_CHAT_HISTORY_FILE = 'project.history'
CONFIG_TASK_HISTORY_FILE = 'task.history'

# Database
DB_NAME = 'database'
TABLE_PROJECTS = 'projects.json'

# LLMs
LLM_TYPE_OPENAI = 'openai'
LLM_TYPE_OLLAMA = 'ollama'

# Search Engine
SEARCH_ENGINE_GOOGLE = 'google'
SEARCH_ENGINE_SEARCHAPI = 'searchapi'
SEARCH_ENGINE_BING = 'bing'

# Integrations
INTEGRATIOIN_AWS_S3 = 'aws_s3'
INTEGRATIOIN_HUGGINGFACE = 'huggingface'
INTEGRATIOIN_SNOWFLAKE = 'snowflake'
INTEGRATION_LIST = [
    INTEGRATIOIN_AWS_S3,
    INTEGRATIOIN_HUGGINGFACE,
    INTEGRATIOIN_SNOWFLAKE
]



import os

import yaml

from agent.types import Task


def load_yml(file_name: str):
    """
    Load a step from a .yaml file.
    :param file_name: the name of the configuration file.
    :return:
    """

    dir_path = os.path.dirname(os.path.realpath(__file__))
    yml_path = os.path.join(dir_path, file_name)

    with open(yml_path, 'r') as file:
        data = yaml.safe_load(file)

    return data


def match_plan(task_dict: dict):
    """
    Match a plan task name with the task description.
    :param task_dict: the dictionary of tasks.
    :return: a task object.
    """

    tasks = load_yml('plan.yml')
    for task in tasks:
        if task_dict.get('name').lower() == task['name'].lower():
            resource_list = []
            if task.get('resources'):
                for resource in task['resources']:
                    if task_dict.get('resources'):
                        if resource['name'] in task_dict.get('resources'):
                            resource_list.append(resource)

            task['resources'] = resource_list
            task['description'] = task_dict.get('description')
            return Task(**task)

    return None



#!/usr/bin/env python
from setuptools import find_packages, setup

# read the contents of README file
from os import path
from io import open  # for Python 2 and 3 compatibility

# get __version__ from _version.py
ver_file = path.join('agent', 'version.py')
with open(ver_file) as f:
    exec(f.read())

this_directory = path.abspath(path.dirname(__file__))


# read the contents of README.md
def readme():
    with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        return f.read()


# read the contents of requirements.txt
with open(path.join(this_directory, 'requirements.txt'), encoding='utf-8') as f:
    requirements = f.read().splitlines()

setup(
    name='mle-agent',
    version=__version__,
    description='MLE-agent: An agent to automate your MLE processes',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='Yizheng Huang, Huaizheng Zhang',
    author_email='huangyz0918@gmail.com',
    url='https://github.com/MLSysOps/MLE-agent',
    download_url='https://github.com/MLSysOps/MLE-agent/archive/refs/heads/main.zip',
    keywords=['LLM', 'deep learning', 'MLOps', 'shell', 'neural networks'],
    packages=find_packages(),
    package_data={
        'metadata': ['agent/hub/*.yml']
    },
    entry_points={
        "console_scripts": [
            "mle-agent=agent.cli:cli",
            "mle=agent.cli:cli",
        ]
    },
    zip_safe=False,
    include_package_data=True,
    install_requires=requirements,
    setup_requires=['setuptools>=38.6.0'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        "Operating System :: OS Independent",
    ],
)



import os
from rich import print
from rich.console import Console
from prettytable import PrettyTable
from colorama import Fore, Style, init

init(autoreset=True)


def get_terminal_size():
    """
    Get the width of the terminal

    Returns:
    int: The width of the terminal
    """
    try:
        columns = os.get_terminal_size().columns
    except OSError:
        columns = 120  # default width if terminal size cannot be determined
    return columns


def split_text(text, max_width=80, max_lines=5):
    """
    Split text based on maximum width and maximum lines

    Args:
    text (str): The text to be split
    max_width (int, optional): Maximum width per line. Defaults to 80.
    max_lines (int, optional): Maximum number of lines. Defaults to 5.

    Returns:
    list: List of split text
    """
    words = text.split()
    lines = []
    current_line = words[0]

    for word in words[1:]:
        if len(current_line) + len(word) < max_width:
            current_line += " " + word
        else:
            if len(lines) >= max_lines - 1:
                current_line += "..."
                break
            lines.append(current_line)
            current_line = word
    lines.append(current_line)
    return lines


def generate_one_plan_card_ascii(
        step: int,
        name: str,
        resources: list[str],
        description: str,
        max_width: int = 80,
        require_arrow: bool = True,
):
    """
    Generate ASCII art for a single plan card

    Args:
        step (int): The step number
        name (str): The name of the task
        resources (list): List of resources
        description (str): Description of the task
        max_width (int, optional): Maximum width. Defaults to 80.
        require_arrow (bool, optional): Whether an arrow is required. Defaults to True.

    Returns:
        str: ASCII art string
    """
    max_width = min(get_terminal_size() // 2 - 10, max_width)
    table = PrettyTable()
    table.title = (
            Fore.YELLOW
            + Style.BRIGHT
            + f"Task {step + 1}: {name}"
            + Style.RESET_ALL
    )
    table.field_names = [
        Fore.BLUE + "Resources" + Style.RESET_ALL,
        Fore.GREEN + ", ".join(resources) + Style.RESET_ALL,
    ]
    table.add_rows(
        [
            [
                (Fore.BLUE + "Description" + Style.RESET_ALL) if i == 0 else "",
                Fore.GREEN + s + Style.RESET_ALL,
            ]
            for i, s in enumerate(split_text(description, max_width))
        ]
    )
    table.hrules = False
    table_string = table.get_string()

    terminal_width = get_terminal_size()
    table_width = len(table_string.split("\n")[0])
    table_padding = (
        (terminal_width - table_width) // 2 if terminal_width > table_width else 0
    )

    strs = ""
    for line in table_string.split("\n"):
        strs += " " * table_padding + line + "\n"

    if require_arrow:
        strs += " " * (terminal_width // 2) + Fore.MAGENTA + "|" + Style.RESET_ALL + "\n"
        strs += " " * (terminal_width // 2) + Fore.MAGENTA + "V" + Style.RESET_ALL + "\n"

    return strs


def generate_plan_card_ascii(task_dicts):
    """
    Generate ASCII art for plan cards

    Args:
        task_dicts (dict): Dictionary of tasks

    Returns:
        str: ASCII art string
    """
    strs = ""
    for i, plan in enumerate(task_dicts["tasks"]):
        strs += generate_one_plan_card_ascii(step=i, require_arrow=i != len(task_dicts["tasks"]) - 1, **plan)
    return strs


def show_panel(text: str):
    """
    Display the text in a panel

    Args:
    text (str): The text to be displayed
    """
    console = Console()
    console.rule()
    print(f"[bold]{text}[bold]")
    console.rule()
